#include "mlp_priv.h"
#include <stdlib.h> 	/* for malloc, strtod, strtol */
#include <string.h> 	/* for strdup */
#include <stdio.h> 		/* for snprintf */



static mlp_value *mlp_value_new(mlp_type t){
	mlp_value *v = (mlp_value *)malloc(sizeof(mlp_value)) ;
	v->t = t ;
	v->stringy = NULL ;

	return v ;
}


/*
	Allocates a new mlp_value object from a long integer.
*/
MLP_PUBLIC_API mlp_value *mlp_value_new_from_long(long val){
	mlp_value *v = mlp_value_new(MLP_LONG) ;
	v->v.l = val ;
	return v ;
}


/*
 	Allocates a new mlp_value object from an integer.
*/
MLP_PUBLIC_API mlp_value *mlp_value_new_from_int(int val){
	mlp_value *v = mlp_value_new(MLP_INT) ;
	v->v.i = val ;
	return v ;
}


/*
 	Allocates a new mlp_value object from from a double.
*/
MLP_PUBLIC_API mlp_value *mlp_value_new_from_double(double val){
	mlp_value *v = mlp_value_new(MLP_DOUBLE) ;
	v->v.d = val ;
	return v ;
}


/*
 	Allocates a new mlp_value object from a char *, which may be NULL.
*/
MLP_PUBLIC_API mlp_value *mlp_value_new_from_string(const char *val){
	mlp_value *v = mlp_value_new(MLP_STRING) ;
	v->v.s = (val == NULL ? NULL : strdup(val)) ;
	return v ;
}


/*
 	Deletes an mlp_value object.
*/
MLP_PUBLIC_API void mlp_value_delete(mlp_value *v){
	switch (v->t){
		case MLP_STRING:
			if (v->v.s != NULL){
				free(v->v.s) ;
			}
			break ;
    }
	if (v->stringy != NULL){
		free(v->stringy) ;
	}
	free(v) ;
}


/*
 	Extracts a long integer from an mlp_value object.
*/
MLP_PUBLIC_API long mlp_value_to_long(mlp_value *v){
	switch (v->t){
		case MLP_INT:
			return (long)v->v.i ;
		case MLP_LONG:
			return v->v.l ;
		case MLP_DOUBLE:
			return (long)v->v.d ;
		case MLP_STRING:
			return (v->v.s == NULL ? strtol("NAN", NULL, 10) : strtol(v->v.s, NULL, 10)) ;
		default:
			return strtol("NAN", NULL, 10) ;
	}
}


/*
 	Extracts an integer from an mlp_value object.
*/
MLP_PUBLIC_API int mlp_value_to_int(mlp_value *v){
	return (int)mlp_value_to_long(v) ;
}


/*
 	Extracts a double from an mlp_value object.
*/
MLP_PUBLIC_API double mlp_value_to_double(mlp_value *v){
	switch (v->t){
		case MLP_INT:
			return (double)v->v.i ;
		case MLP_LONG:
			return (double)v->v.l ;
		case MLP_DOUBLE:
			return v->v.d ;
		case MLP_STRING:
			return (v->v.s == NULL ? strtod("NAN", NULL) : strtod(v->v.s, NULL)) ;
		default:
			return strtod("NAN", NULL) ;
	}
}


/*
 	Extracts a char *. from an mlp_value object. 
	Note: The returned char * is kept in an internal buffer and therefore must not be freed by the caller.
*/
MLP_PUBLIC_API const char *mlp_value_to_string(mlp_value *v){
	if (v->stringy != NULL){
		return v->stringy ;
	}

	switch (v->t){
		case MLP_STRING:
			return v->v.s ;
		case MLP_INT: {
			char num[1024] ;
			snprintf(num, 1023, "%d", v->v.i) ;
			v->stringy = strdup(num) ;
			return v->stringy ;
		}
		case MLP_LONG: {
			char num[1024] ;
			snprintf(num, 1023, "%ld",v->v.l) ;
			v->stringy = strdup(num) ;
			return v->stringy ;
		}
		case MLP_DOUBLE: {
			char num[1024] ;
			snprintf(num, 1023, "%.15lf", v->v.d) ;
			v->stringy = strdup(num) ;
			return v->stringy ;
		}
		default:
			return NULL ;
	}
}


/*
 	Returns the type of an mlp_value object.
*/
MLP_PUBLIC_API mlp_type mlp_value_get_type(mlp_value *v){
	return v->t ;
}
