#ifndef _MLP_HPP_
#define _MLP_HPP_


extern "C" {
	#include "mlp.h"
}


#define MLP_API extern "C"


#include <string>
#include <map>
#include <vector>


class mlp_language ;
class mlp_runtime ;
class mlp_library ;
class mlp_function ;
class mlp_array ;
class mlp_object ;
typedef std::vector<mlp_type> mlp_prototype ;
typedef std::vector<mlp_value> mlp_arguments ;


/* MLP Context */
class mlp_context {
	mlp_error err ;
	char *err_msg ;
	std::map<mlp_runtime *, const void *> runtime_data ;

	public:
		mlp_context() ;
		int has_error() ;
		mlp_error get_error() ;
		const char *get_error_msg() ;
		void set_verror(mlp_error e, const char *fmt, va_list) ;
		void set_error(mlp_error e, const char *fmt, ...) ;
		void clear_error() ;
		void set_runtime_data(mlp_runtime *rt, const void *data) ;
		const void *get_runtime_data(mlp_runtime *rt) ;
		~mlp_context() ;
} ;


/* MLP Value */
class mlp_value {
	mlp_type t ;
	union {
		int i ;
		long l ;
		double d ;
		char *s ;
		const mlp_array *a ;
		const mlp_object *o ;
	} v ;
	bool thrown ;
	char *stringy ;

	private:
		mlp_value(mlp_type t) ;
	public:
		mlp_value(const mlp_value &v) ;
		mlp_value(int val) ;
		mlp_value(long val) ;
		mlp_value(double val) ;
		mlp_value(const char *val) ;
		mlp_value(const mlp_object *val) ;
		mlp_value(const mlp_array *val) ;
		int to_int() ;
		long to_long() ;
		double to_double() ;
		const char *to_string() ;
		const mlp_object *to_object() ;
		const mlp_array *to_array() ;
		mlp_type get_type() ;
		virtual ~mlp_value() ;
} ;


/* MLP Language */
class mlp_language {
	std::string name ;
	static std::map<std::string, mlp_language *> languages ;
	void *handle ;

	protected: 
		mlp_language(const char *name) ;
		~mlp_language() ;
	private:
		void set_handle(void *handle) ;
	public: 
		static mlp_language *init(mlp_context *ctx, const char *name) ;
		virtual mlp_runtime *new_runtime(mlp_context *ctx, const void *opts) = 0 ;
		// virtual runtime *attach_runtime(context *ctx, const void *) = 0 ;
		const char *get_name(mlp_context *ctx) ;
		void destroy(mlp_context *ctx) ;
} ;


/* MLP Runtime */
class mlp_runtime {
	mlp_language *lang ;

	protected:
		mlp_runtime(mlp_context *ctx, mlp_language *lang, const void *opts) ;
		virtual ~mlp_runtime() ;
	public:
		virtual mlp_library *load_library(mlp_context *ctx, const char *name, const void *opts) = 0 ;
		virtual void setup_context(mlp_context *ctx) = 0 ;
		//virtual value *call_function(context *ctx, const char *name, int nb_args, const value **args) = 0 ;
		//virtual object *new_object(context *ctx, const clazz *cls, int nb_args, const value **args) = 0 ;
		//virtual mlp_array *new_array(mlp_context *ctx, mlp_type type, mlp_arguments *args) = 0 ;
		mlp_language *get_language(mlp_context *ctx) ;
		virtual void destroy(mlp_context *ctx) ;
} ;


/* MLP Library */
class mlp_library {
	protected:
	std::string name ;
	mlp_runtime *rt ;

	protected:
		mlp_library(mlp_context *ctx, mlp_runtime *rt, const char *name, const void *opts) ;
		virtual ~mlp_library() ;
	public:
		virtual mlp_function *get_function(mlp_context *ctx, 
			mlp_type rtype, const char *name, mlp_prototype proto) = 0 ;
		const char *get_name(mlp_context *ctx) ;
		mlp_runtime *get_runtime(mlp_context *ctx) ;
		virtual void destroy(mlp_context *ctx) ;
} ;


/* MLP Function */
class mlp_function {
	protected:
	std::string name ;
	mlp_runtime *rt ;
	mlp_library *lib ;
	mlp_type rtype ;
	mlp_prototype proto ;

	protected:
		mlp_function(mlp_context *ctx, mlp_runtime *rt, mlp_library *lib, 
			mlp_type rtype, const char *name, mlp_prototype proto) ;
		virtual ~mlp_function() ;
	public:
		virtual mlp_value *call(mlp_context *ctx, mlp_arguments args) = 0 ;
		const char *get_name(mlp_context *ctx) ;
		mlp_runtime *get_runtime(mlp_context *ctx) ;
		mlp_library *get_library(mlp_context *ctx) ;
		mlp_type get_return_type(mlp_context *ctx) ;
		virtual void destroy(mlp_context *ctx) ;
} ;




/* MLP Array */
class mlp_array {
	mlp_type t ;
	int len ;

	protected:
	public:
	int get_length(mlp_context *ctx) ;
	mlp_value *get_element(mlp_context *ctx, int idx) ;
	void *set_element(mlp_context *ctx, int idx, mlp_value *v) ;
} ;


/* A class */
class mlp_class {
	char *name ;

	public:
	const char *get_name() ;
} ;


/* An object */
class mlp_object {
	mlp_class *cls ;

	public:
	mlp_class *get_class() ;
} ;


#endif
