#ifndef _MLP_H_
#define _MLP_H_

#include <stdarg.h>


#define MLP_VERSION "0.01"


/* MLP error codes */
enum _mlp_error {
	MLP_NOT_SUPPORTED = -1000,  	/* Language does not support this feature */
	MLP_NOT_IMPLEMENTED = -1001, 	/* Language supports this feature but plugin doesn't implement it yet... */

	MLP_OK = 0,
	MLP_PLUGIN_ERROR,
	MLP_LANGUAGE_ERROR,
	MLP_LIBRARY_ERROR,
	MLP_RUNTIME_ERROR,
	MLP_FUNCTION_ERROR,

	MLP_NO_SUCH_FUNCTION,
	MLP_NO_SUCH_CLASS,
	MLP_INVALID_ARGUMENT,
	MLP_INVALID_RETURN_VALUE,
	MLP_INTERNAL_ERROR
} ;
typedef enum _mlp_error mlp_error ;


/* MLP supported data types */
enum _mlp_type {
	MLP_INT,
	MLP_LONG,
	MLP_DOUBLE,
	MLP_STRING,
	MLP_ARRAY,
	MLP_OBJECT
} ;
typedef enum _mlp_type mlp_type ;



#ifdef __cplusplus


#include <string>
#include <map>
#include <vector>


#define MLP_API extern "C"
class mlp_context ;
class mlp_language ;
class mlp_runtime ;
class mlp_library ;
class mlp_function ;
class mlp_array ;
class mlp_object ;
class mlp_value ;
typedef std::vector<mlp_type> mlp_prototype ;
typedef std::vector<mlp_value> mlp_arguments ;


/* MLP Context */
class mlp_context {
	mlp_error err ;
	char *err_msg ;
	std::map<mlp_runtime *, const void *> runtime_data ;

	public:
	mlp_context() ;
	int has_error() ;
	mlp_error get_error() ;
	const char *get_error_msg() ;
	void set_verror(mlp_error e, const char *fmt, va_list) ;
	void set_error(mlp_error e, const char *fmt, ...) ;
	void clear_error() ;
	void set_runtime_data(mlp_runtime *rt, const void *data) ;
	const void *get_runtime_data(mlp_runtime *rt) ;
	~mlp_context() ;
} ;


/* MLP Value */
class mlp_value {
	mlp_type type ;
	union {
		int i ;
		long l ;
		double d ;
		char *s ;
		const mlp_array *a ;
		const mlp_object *o ;
	} v ;
	bool thrown ;
	char *stringy ;

	private:
	mlp_value(mlp_type t) ;
	public:
	mlp_value(const mlp_value &v) ;
	mlp_value(int val) ;
	mlp_value(long val) ;
	mlp_value(double val) ;
	mlp_value(const char *val) ;
	mlp_value(const mlp_object *val) ;
	mlp_value(const mlp_array *val) ;
	int to_int() ;
	long to_long() ;
	double to_double() ;
	const char *to_string() ;
	const mlp_object *to_object() ;
	const mlp_array *to_array() ;
	mlp_type get_type() ;
	virtual ~mlp_value() ;
} ;


/* MLP Language */
class mlp_language_impl ;
class mlp_language {
	mlp_language_impl *impl ;

	private: 
		mlp_language(mlp_language_impl *impl) ;
		~mlp_language() ; 
	public: 
		static mlp_language *init(mlp_context *ctx, const char *name) ;
		const char *get_name(mlp_context *ctx) ;
		mlp_runtime *new_runtime(mlp_context *ctx, const void *opts) ; 
		bool destroy(mlp_context *ctx) ;
} ;


/* MLP Runtime */
class mlp_runtime_impl ;
class mlp_runtime {
	friend class mlp_language ;
	mlp_runtime_impl *impl ;
	mlp_language *lang ;

	private: 
		mlp_runtime(mlp_runtime_impl *impl) ;
		~mlp_runtime() ; 
} ;


#else


#define MLP_API
struct _mlp_context ;
typedef struct _mlp_context mlp_context ;
struct _mlp_value ;
typedef struct _mlp_value mlp_value ;
struct _mlp_language ;
typedef struct _mlp_language mlp_language ;
struct _mlp_runtime ;
typedef struct _mlp_runtime mlp_runtime ;
struct _mlp_library ;
typedef struct _mlp_library mlp_library ;
struct _mlp_function ;
typedef struct _mlp_function mlp_function ;
struct _mlp_array ;
typedef struct _mlp_array mlp_array ;
struct _mlp_object ;
typedef struct _mlp_object mlp_object ;


#endif

#include "mlp_p.h"

#endif
