package mlp.java ;


import java.util.* ;
import mlp.* ;


class MLPJavaUtils {

	static Class find(String name) throws MLPJavaException {
		try {
			Class c = Class.forName(name) ;
			return box(c) ;
		}
		catch (ClassNotFoundException e){
			throw new MLPJavaException("Class " + name + " not found") ;
		}
	}


	static boolean canCast(Class<?> proto[], MLPValue args[]){
		MLPUtils.debug(3, "checking if " + makeSignature(args) + " is compatible with " + makeSignature(proto)) ;
		int plen = (proto != null ? proto.length : 0) ;
		int alen = (args != null ? args.length : 0) ;
		if (plen != alen){
			MLPUtils.debug(4, "lengths are different: " + proto.length + " != " + args.length) ;
			return false ;
		}

		for (int i = 0 ; i < plen ; i++){	
			Class p = proto[i] ;
			MLPValue a = args[i] ;
	
			if (! canCast(p, a)){
				MLPUtils.debug(4, "halting") ;
				return false ;
			}
		}

		return true ;
	}


	/* 
	 *  Verify if an object of type p can be made from MLPValue a.
	 */
	static boolean canCast(Class<?> p, MLPValue a){
		MLPUtils.debug(4, "arg   is of type " + a.getType().getName()) ;
		MLPUtils.debug(4, "proto is of type " + p.getName()) ;

		if (p.equals(java.lang.Object.class)){
			return true ;
		}

		// Box the type if it is a primitive type.
		p = box(p) ;

		MLPType atype = a.getType() ;
		String details = "" ;
		switch (atype){
			case BOOL: return isBool(p) ; 
			case LONG: return isIntegral(p) ;
			case DOUBLE: return isFloating(p) ;
			case CHAR: return isChar(p) ;
			case STRING: return isString(p) ;
			case OBJECT: {
				if (isReference(p)){
					MLPObject o = a.getObject() ;
					if (o == null){
						return true ;
					}
					else {
						Class type = null ;
						if (o instanceof MLPJavaObject){
							// Java object
							type = ((MLPJavaObject)o).getObject().getClass() ;
						}
						else {
							// Some other language object
							type = o.getClass() ;
						}
						if (p.isAssignableFrom(type)){
							return true ;
						}
						else {
							details = "(" + type.getName() + ")" ;
						}
					}
				}
				break ;
			}
		}

		// If we reach here, it means there is some sort of incompatibility.
		MLPUtils.debug(3, "Can't cast arg of type " + a.getType().getName() + details + " to proto type " + p.getName()) ;

		return false ;
	}


	/* 
	 * Try to fit the arguments to the prototype.
	 * Here it is assumed that proto[] and args[] are of the same length.
	 */
	static Object [] castArguments(Class proto[], MLPValue args[]) throws MLPJavaException {
		MLPUtils.debug(3, "casting " + makeSignature(args) + " to " + makeSignature(proto)) ;
		int plen = (proto != null ? proto.length : 0) ;
		int alen = (args != null ? args.length : 0) ;
		Object ret[] = new Object[plen] ;
	
		for (int i = 0 ; i < plen ; i++){	
			Class p = proto[i] ;
			MLPValue a = args[i] ;

			ret[i] = castArgument(p, a) ;
		}

		return ret ;
	}


	static Object castArgument(Class<?> p, MLPValue a) throws MLPJavaException {
		MLPUtils.debug(4, "proto is of type " + p.getName()) ;
		MLPUtils.debug(4, "arg   is of type " + a.getType().getName()) ;

		// Box the type if it is a primitive type.
		p = box(p) ; 

		MLPType atype = a.getType() ;
		String details=  "" ;
		switch (atype){
			case BOOL: {
				boolean b = a.getBool() ;
				if ((p.equals(java.lang.Boolean.class))
					||(p.equals(java.lang.Object.class))){
					return new Boolean(b) ; 
				}
				break ;
			}
			case LONG: {
				long l = a.getLong() ;
				if (p.equals(java.lang.Byte.class)){
					return new Byte((byte)l) ; 
				}
				else if (p.equals(java.lang.Short.class)){
					return new Short((short)l) ; 
				}
				else if (p.equals(java.lang.Integer.class)){
					return new Integer((int)l) ; 
				}
				else if ((p.equals(java.lang.Long.class))
						||(p.equals(java.lang.Number.class))
						||(p.equals(java.lang.Object.class))){
					return new Long(l) ; 
				}
				break ;
			}
			case DOUBLE: {
				double d = a.getDouble() ;
				if (p.equals(java.lang.Float.class)){
					return new Float((float)d) ; 
				}
				else if ((p.equals(java.lang.Double.class))
						||(p.equals(java.lang.Number.class))
						||(p.equals(java.lang.Object.class))){
					return new Double((double)d) ; 
				}
				break ;
			}
			case CHAR: {
				char c = a.getChar() ;
				if ((p.equals(java.lang.Character.class))
					||(p.equals(java.lang.Object.class))){
					return new Character(c) ; 
				}
				break ;
			}
			case STRING: {
				String s = a.getString() ;
				if (s == null){
					return null ;
				}
				else if ((p.equals(java.lang.String.class))
						||(p.equals(java.lang.Object.class))){
					return new String(s) ; 
				}
				break ;
			}
			case OBJECT: {
				MLPObject o = a.getObject() ;
				if (o == null){
					return null ;
				}
				else {
					if (o instanceof MLPJavaObject){
						// Java object
						Class type = ((MLPJavaObject)o).getObject().getClass() ;
						if (p.isAssignableFrom(type)){
							return ((MLPJavaObject)o).getObject() ;
						}
						else {
							details = "(" + type.getName() + ")" ;
						}
					}
					else {
						// Some other language object
						Class type = o.getClass() ;
						if (p.isAssignableFrom(type)){
							return o ;
						}
						else {
							details = "(" + type.getName() + ")" ;
						}
					}
				}
				break ;
			}
		}

		// If we reach here, it means there is some sort of incompatibility.
		throw new MLPJavaCastException("Can't cast arg of type " + a.getType().getName() + details + " to proto type " + p.getName()) ;
	}


	static MLPValue wrapObject(Object obj, Class p){
		p = box(p) ;
		//box_classes.put(char.class, java.lang.Character.class) ;
		if (p.equals(java.lang.Boolean.class)){
			return new MLPValue(((Boolean)obj).booleanValue()) ;
		}
		else if (p.equals(java.lang.Byte.class)){
			return new MLPValue((long)((Byte)obj).byteValue()) ;
		}
		else if (p.equals(java.lang.Short.class)){
			return new MLPValue((long)((Short)obj).shortValue()) ;
		}
		else if (p.equals(java.lang.Integer.class)){
			return new MLPValue((long)((Integer)obj).intValue()) ;
		}
		else if (p.equals(java.lang.Long.class)){
			return new MLPValue((long)((Long)obj).longValue()) ;
		}
		else if (p.equals(java.lang.Float.class)){
			return new MLPValue((double)((Float)obj).floatValue()) ;
		}
		else if (p.equals(java.lang.Double.class)){
			return new MLPValue((double)((Double)obj).doubleValue()) ;
		}
		else if (p.equals(java.lang.Character.class)){
			return new MLPValue((char)((Character)obj).charValue()) ;
		}
		else if (p.equals(java.lang.String.class)){
			return new MLPValue((String)obj) ;
		}
		else {
			return new MLPValue(new MLPJavaObject(obj, p)) ;
		}
	}


	static String makeSignature(Class proto[]){
        StringBuffer ret = new StringBuffer() ;
		int plen = (proto != null ? proto.length : 0) ;
        for (int i = 0 ; i < plen ; i++){
            if (i > 0){
                ret.append(",") ;
            }
            ret.append(proto[i].getName()) ;
        }

        return "(" + ret.toString() + ")" ;
	}

	static String makeSignature(MLPValue args[]){
        StringBuffer ret = new StringBuffer() ;
		int alen = (args != null ? args.length : 0) ;
        for (int i = 0 ; i < alen ; i++){
            if (i > 0){
                ret.append(",") ;
            }
            ret.append(args[i].getType().getName()) ;
        }

        return "(" + ret.toString() + ")" ;
	}

	static String makeSignature(String proto[]){
        StringBuffer ret = new StringBuffer() ;
		int plen = (proto != null ? proto.length : 0) ;
        for (int i = 0 ; i < plen ; i++){
            if (i > 0){
                ret.append(",") ;
            }
            ret.append(proto[i]) ;
        }

        return "(" + ret.toString() + ")" ;
	}


    /*
	 *  Finds the appropriate boxing type, if required.
	 */
    static private HashMap<Class,Class> box_classes = new HashMap<Class,Class>() ;
    static {
		box_classes.put(byte.class, java.lang.Byte.class) ;
		box_classes.put(short.class, java.lang.Short.class) ;
		box_classes.put(int.class, java.lang.Integer.class) ;
		box_classes.put(long.class, java.lang.Long.class) ;
		box_classes.put(float.class, java.lang.Float.class) ;
		box_classes.put(double.class, java.lang.Double.class) ;
		box_classes.put(boolean.class, java.lang.Boolean.class) ;
		box_classes.put(char.class, java.lang.Character.class) ;
		box_classes.put(void.class, java.lang.Void.class) ;
    } ;
	static Class box(Class p){
		Class box = box_classes.get(p) ;

		return (box != null ? box : p) ;
	}


	static private HashMap<Class,Boolean> integral_classes = new HashMap<Class,Boolean>() ;
	static {
		integral_classes.put(java.lang.Byte.class, new Boolean(true)) ;
		integral_classes.put(java.lang.Short.class, new Boolean(true)) ;
		integral_classes.put(java.lang.Integer.class, new Boolean(true)) ;
		integral_classes.put(java.lang.Long.class, new Boolean(true)) ;
		integral_classes.put(java.lang.Number.class, new Boolean(true)) ;
	} ;
	static boolean isIntegral (Class p){
		return (integral_classes.get(p) != null) ;
	}


	static private HashMap<Class,Boolean> floating_classes = new HashMap<Class,Boolean>() ;
	static {
		floating_classes.put(java.lang.Float.class, new Boolean(true)) ;
		floating_classes.put(java.lang.Double.class, new Boolean(true)) ;
		floating_classes.put(java.lang.Number.class, new Boolean(true)) ;
	} ;
	static boolean isFloating (Class p){
		return (floating_classes.get(p) != null) ;
	}


	static boolean isNumeric (Class p){
		return (isIntegral(p))||(isFloating(p)) ;
	}


	/*
	 * Class is String
	*/
	static private HashMap<Class,Boolean> string_classes = new HashMap<Class,Boolean>() ;
	static {
		string_classes.put(java.lang.String.class, new Boolean(true)) ;
	}
	static boolean isString (Class p){
		return (string_classes.get(p) != null) ;
	}


	/*
		Class is Char
	*/
	static private HashMap<Class,Boolean> char_classes = new HashMap<Class,Boolean>() ;
	static {
		char_classes.put(java.lang.Character.class, new Boolean(true)) ;
	}
	static boolean isChar (Class p){
		return (char_classes.get(p) != null) ;
	}


	/*
	 * Class is Bool
	*/
	static private HashMap<Class,Boolean> bool_classes = new HashMap<Class,Boolean>() ;
	static {
		bool_classes.put(java.lang.Boolean.class, new Boolean(true)) ;
	}
	static boolean isBool (Class p){
		return (bool_classes.get(p) != null) ;
	}


	static boolean isPrimitive(Class p){
		if ((isNumeric(p))||(isString(p))||(isChar(p))||(isBool(p))){
			String name = p.getName() ;
			MLPUtils.debug(5, "class " + name + " is primitive") ;
			return true ;
		}

		return false ;
	}

	
	/*
		Determines if a class is not of a primitive type.
	*/
	static boolean isReference(Class p){
		if (isPrimitive(p)){
			return false ;
		}
		else {
			String name = p.getName() ;
			MLPUtils.debug(5, "class " + name + " is reference") ;
		}

		return true ;
	}


	/*
	static boolean ClassIsArray (Class p){
		if ((isReference(p))&&(name.startsWith("["))){
			String name = p.getName() ;
			MLPUtils.debug(4, "class " + name + " is array") ;
			return true ;
		}

		return false ;
	}
	*/


	static boolean isPublic(Class p){
		int pub = p.getModifiers() & java.lang.reflect.Modifier.PUBLIC ;
		if (pub != 0){
			return true ;
		}

		return false ;
	}
}
